==> optimum_neuron-0.0.16-py3-none-any.whl: optimum/neuron/modeling_decoder.py <==
# coding=utf-8
# Copyright 2023 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for text-generation model architectures on neuron devices."""

import logging
import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Optional, Tuple, Union

import torch
from huggingface_hub import HfApi, HfFolder, snapshot_download
from huggingface_hub.utils import is_google_colab
from transformers import AutoConfig, AutoModel, GenerationConfig

from ..exporters.neuron.model_configs import *  # noqa: F403
from ..exporters.tasks import TasksManager
from ..modeling_base import OptimizedModel
from .utils import is_transformers_neuronx_available
from .utils.version_utils import check_compiler_compatibility, get_neuronxcc_version


if is_transformers_neuronx_available():
    from transformers_neuronx.module import PretrainedModel as NeuronxPretrainedModel
    from transformers_neuronx.module import save_split


if TYPE_CHECKING:
    from transformers import PretrainedConfig


logger = logging.getLogger(__name__)


def get_exporter(config, task):
    return TasksManager.get_exporter_config_constructor(model_type=config.model_type, exporter="neuron", task=task)()


class NeuronDecoderModel(OptimizedModel):
    """
    Base class to convert and run pre-trained transformers decoder models on Neuron devices.

    It implements the methods to convert a pre-trained transformers decoder model into a Neuron transformer model by:
    - transferring the checkpoint weights of the original into an optimized neuron graph,
    - compiling the resulting graph using the Neuron compiler.

    Common attributes:
        - model (`torch.nn.Module`) -- The decoder model with a graph optimized for neuron devices.
        - config ([`~transformers.PretrainedConfig`]) -- The configuration of the original model.
        - generation_config ([`~transformers.GenerationConfig`]) -- The generation configuration used by default when calling `generate()`.
    """

    model_type = "neuron_model"
    auto_model_class = AutoModel

    CHECKPOINT_DIR = "checkpoint"
    COMPILED_DIR = "compiled"

    def __init__(
        self,
        model: torch.nn.Module,
        config: "PretrainedConfig",
        model_path: Union[str, Path, TemporaryDirectory],
        generation_config: Optional[GenerationConfig] = None,
    ):
        if not is_transformers_neuronx_available() or not isinstance(model, NeuronxPretrainedModel):
            raise ValueError("The source model must be a transformers_neuronx.PreTrainedModel.")

        super().__init__(model, config)
        self.model_path = model_path
        if generation_config is None:
            generation_config = GenerationConfig.from_model_config(config)
        self.generation_config = generation_config
        # Registers the NeuronModelForXXX classes into the transformers AutoModel classes to avoid warnings when creating
        # a pipeline https://github.com/huggingface/transformers/blob/3d3204c025b6b5de013e07dd364208e28b4d9589/src/transformers/pipelines/base.py#L940
        AutoConfig.register(self.model_type, AutoConfig)
        if hasattr(self.auto_model_class, "register"):
            self.auto_model_class.register(AutoConfig, self.__class__)

    @classmethod
    def _from_transformers(
        cls,
        model_id: str,
        config: "PretrainedConfig",
        use_auth_token: Optional[Union[bool, str]] = None,
        revision: Optional[str] = None,
        force_download: bool = False,
        cache_dir: Optional[str] = None,
        subfolder: str = "",
        local_files_only: bool = False,
        trust_remote_code: bool = False,
        task: Optional[str] = None,
        batch_size: Optional[int] = 1,
        sequence_length: Optional[int] = None,
        num_cores: Optional[int] = 2,
        auto_cast_type: Optional[str] = "fp32",
        **kwargs,
    ) -> "NeuronDecoderModel":
        if not is_transformers_neuronx_available():
            raise ModuleNotFoundError("The transformers_neuronx package is required to export the model.")

        if task is None:
            task = TasksManager.infer_task_from_model(cls.auto_model_class)

        # Instantiate the transformers model checkpoint
        model = TasksManager.get_model_from_task(
            task=task,
            model_name_or_path=model_id,
            subfolder=subfolder,
            revision=revision,
            framework="pt",
            cache_dir=cache_dir,
            use_auth_token=use_auth_token,
            local_files_only=local_files_only,
            force_download=force_download,
            trust_remote_code=trust_remote_code,
            **kwargs,
        )

        # Save the model checkpoint in a temporary directory
        checkpoint_dir = TemporaryDirectory()
        model.save_pretrained(
            checkpoint_dir.name, save_function=save_split, safe_serialization=False, max_shard_size="10000GB"
        )

        # If the sequence_length was not specified, deduce it from the model configuration
        if sequence_length is None:
            # Note: for older models, max_position_embeddings is an alias for n_positions
            sequence_length = config.max_position_embeddings

        # Update the config
        config.neuron = {
            "task": task,
            "batch_size": batch_size,
            "num_cores": num_cores,
            "auto_cast_type": auto_cast_type,
            "sequence_length": sequence_length,
            "compiler_type": "neuronx-cc",
            "compiler_version": get_neuronxcc_version(),
        }

        return cls._from_pretrained(checkpoint_dir, config)

    @classmethod
    def _get_neuron_paths(
        cls, model_dir: Union[str, Path, TemporaryDirectory], token: Optional[str] = None
    ) -> Tuple[str, str, str]:
        if isinstance(model_dir, TemporaryDirectory):
            model_path = model_dir.name
            # We are in the middle of an export: the checkpoint is in the temporary model directory
            checkpoint_path = model_path
            # There are no compiled artifacts yet
            compiled_path = None
        else:
            # The model has already been exported
            if os.path.isdir(model_dir):
                model_path = model_dir
            else:
                # Download the neuron model from the Hub
                model_path = snapshot_download(model_dir, token=token)
            # The checkpoint is in a subdirectory
            checkpoint_path = os.path.join(model_path, cls.CHECKPOINT_DIR)
            # So are the compiled artifacts
            compiled_path = os.path.join(model_path, cls.COMPILED_DIR)
        return model_path, checkpoint_path, compiled_path

    @classmethod
    def _from_pretrained(
        cls,
        model_id: Union[str, Path, TemporaryDirectory],
        config: "PretrainedConfig",
        use_auth_token: Optional[str] = None,
        **kwargs,
    ) -> "NeuronDecoderModel":
        # Verify we are actually trying to load a neuron model
        neuron_config = getattr(config, "neuron", None)
        if neuron_config is None:
            raise ValueError(
                "The specified directory does not contain a neuron model. "
                "Please convert your model to neuron format by passing export=True."
            )

        # Evaluate the configuration passed during export
        task = neuron_config["task"]
        batch_size = neuron_config["batch_size"]
        sequence_length = neuron_config["sequence_length"]
        num_cores = neuron_config["num_cores"]
        auto_cast_type = neuron_config["auto_cast_type"]

        check_compiler_compatibility(neuron_config["compiler_type"], neuron_config["compiler_version"])

        exporter = get_exporter(config, task)

        model_path, checkpoint_path, compiled_path = cls._get_neuron_paths(model_id, use_auth_token)

        # transformers-neuronx uses f32/f16 instead of fp32/fp16
        auto_cast_type = auto_cast_type.replace("p", "")
        neuronx_model = exporter.neuronx_class.from_pretrained(
            checkpoint_path,
            batch_size=batch_size,
            n_positions=sequence_length,
            tp_degree=num_cores,
            amp=auto_cast_type,
        )

        if compiled_path is not None:
            # Specify the path where compiled artifacts are stored before conversion
            neuronx_model._load_compiled_artifacts(compiled_path)

        # Compile the Neuron model (if present compiled artifacts will be reloaded instead of compiled)
        neuron_cc_flags = os.environ.get("NEURON_CC_FLAGS", "")
        os.environ["NEURON_CC_FLAGS"] = neuron_cc_flags + " --model-type=transformer"
        neuronx_model.to_neuron()
        os.environ["NEURON_CC_FLAGS"] = neuron_cc_flags

        # Try to reload the generation config (if any)
        generation_config = None
        try:
            generation_config = GenerationConfig.from_pretrained(model_path)
        except OSError:
            logger.info("Generation config file not found, using a generation config created from the model config.")

        return cls(neuronx_model, config, model_id, generation_config)

    def forward(self, *args, **kwargs):
        raise NotImplementedError()

    def _save_pretrained(self, save_directory: Union[str, Path]):
        _, src_chkpt_path, src_compiled_path = self._get_neuron_paths(self.model_path)
        _, dst_chkpt_path, dst_compiled_path = self._get_neuron_paths(save_directory)

        shutil.copytree(src_chkpt_path, dst_chkpt_path)

        if src_compiled_path is None:
            # The compiled model has never been serialized: do it now
            self.model._save_compiled_artifacts(dst_compiled_path)
        else:
            shutil.copytree(src_compiled_path, dst_compiled_path)

        if isinstance(self.model_path, TemporaryDirectory):
            # Let temporary directory go out-of-scope to release disk space
            self.model_path = save_directory

        self.generation_config.save_pretrained(save_directory)

    def push_to_hub(
        self,
        save_directory: str,
        repository_id: str,
        private: Optional[bool] = None,
        revision: Optional[str] = None,
        use_auth_token: Union[bool, str] = True,
        endpoint: Optional[str] = None,
    ) -> str:
        if isinstance(use_auth_token, str):
            huggingface_token = use_auth_token
        elif use_auth_token:
            huggingface_token = HfFolder.get_token()
        else:
            raise ValueError("You need to provide `use_auth_token` to be able to push to the hub")
        api = HfApi(endpoint=endpoint)

        user = api.whoami(huggingface_token)
        if is_google_colab():
            # Only in Google Colab to avoid the warning message
            self.git_config_username_and_email(git_email=user["email"], git_user=user["fullname"])

        api.create_repo(
            token=huggingface_token,
            repo_id=repository_id,
            exist_ok=True,
            private=private,
        )
        api.upload_folder(
            repo_id=repository_id, folder_path=save_directory, token=huggingface_token, revision=revision
        )

==> optimum_neuron-0.0.17-py3-none-any.whl: optimum/neuron/modeling_decoder.py <==
# coding=utf-8
# Copyright 2023 The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base class for text-generation model architectures on neuron devices."""

import logging
import os
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Optional, Tuple, Union

from huggingface_hub import HfApi, get_token, snapshot_download
from huggingface_hub.utils import is_google_colab
from transformers import AutoConfig, AutoModel, GenerationConfig

from ..exporters.neuron.model_configs import *  # noqa: F403
from ..exporters.tasks import TasksManager
from ..modeling_base import OptimizedModel
from .utils import hub_neuronx_cache, is_transformers_neuronx_available
from .utils.require_utils import requires_transformers_neuronx
from .utils.version_utils import check_compiler_compatibility, get_neuronxcc_version


if is_transformers_neuronx_available():
    from transformers_neuronx.module import save_split


if TYPE_CHECKING:
    from transformers import PretrainedConfig


logger = logging.getLogger(__name__)


def get_exporter(config, task):
    return TasksManager.get_exporter_config_constructor(model_type=config.model_type, exporter="neuron", task=task)()


class NeuronDecoderModel(OptimizedModel):
    """
    Base class to convert and run pre-trained transformers decoder models on Neuron devices.

    It implements the methods to convert a pre-trained transformers decoder model into a Neuron transformer model by:
    - transferring the checkpoint weights of the original into an optimized neuron graph,
    - compiling the resulting graph using the Neuron compiler.

    Common attributes:
        - model (`torch.nn.Module`) -- The decoder model with a graph optimized for neuron devices.
        - config ([`~transformers.PretrainedConfig`]) -- The configuration of the original model.
        - generation_config ([`~transformers.GenerationConfig`]) -- The generation configuration used by default when calling `generate()`.
    """

    model_type = "neuron_model"
    auto_model_class = AutoModel

    CHECKPOINT_DIR = "checkpoint"
    COMPILED_DIR = "compiled"

    @requires_transformers_neuronx
    def __init__(
        self,
        config: "PretrainedConfig",
        checkpoint_dir: Union[str, Path, TemporaryDirectory],
        compiled_dir: Optional[Union[str, Path, TemporaryDirectory]] = None,
        generation_config: Optional[GenerationConfig] = None,
    ):
        neuron_config = getattr(config, "neuron", None)
        if neuron_config is None:
            raise ValueError(
                "The specified model is not a neuron model."
                "Please convert your model to neuron format by passing export=True."
            )

        self.checkpoint_dir = checkpoint_dir
        self.compiled_dir = compiled_dir
        if generation_config is None:
            logger.info("Generation config file not found, using a generation config created from the model config.")
            generation_config = GenerationConfig.from_model_config(config)
        self.generation_config = generation_config
        # Registers the NeuronModelForXXX classes into the transformers AutoModel classes to avoid warnings when creating
        # a pipeline https://github.com/huggingface/transformers/blob/3d3204c025b6b5de013e07dd364208e28b4d9589/src/transformers/pipelines/base.py#L940
        AutoConfig.register(self.model_type, AutoConfig)
        if hasattr(self.auto_model_class, "register"):
            self.auto_model_class.register(AutoConfig, self.__class__)

        # Evaluate the configuration passed during export
        task = neuron_config["task"]
        batch_size = neuron_config["batch_size"]
        sequence_length = neuron_config["sequence_length"]
        num_cores = neuron_config["num_cores"]
        auto_cast_type = neuron_config["auto_cast_type"]

        check_compiler_compatibility(neuron_config["compiler_type"], neuron_config["compiler_version"])

        exporter = get_exporter(config, task)

        # transformers-neuronx uses f32/f16 instead of fp32/fp16
        auto_cast_type = auto_cast_type.replace("p", "")
        checkpoint_path = checkpoint_dir.name if isinstance(checkpoint_dir, TemporaryDirectory) else checkpoint_dir
        neuronx_model = exporter.neuronx_class.from_pretrained(
            checkpoint_path,
            batch_size=batch_size,
            n_positions=sequence_length,
            tp_degree=num_cores,
            amp=auto_cast_type,
        )

        if compiled_dir is not None:
            # Specify the path where compiled artifacts are stored before conversion
            neuronx_model.load(compiled_dir)

        # Compile the Neuron model (if present compiled artifacts will be reloaded instead of compiled)
        neuron_cc_flags = os.environ.get("NEURON_CC_FLAGS", "")
        os.environ["NEURON_CC_FLAGS"] = neuron_cc_flags + " --model-type=transformer"
        with hub_neuronx_cache():
            neuronx_model.to_neuron()
        os.environ["NEURON_CC_FLAGS"] = neuron_cc_flags

        super().__init__(neuronx_model, config)

    @classmethod
    def _create_checkpoint(
        cls,
        model_id: str,
        use_auth_token: Optional[Union[bool, str]] = None,
        revision: Optional[str] = None,
        force_download: bool = False,
        cache_dir: Optional[str] = None,
        subfolder: str = "",
        local_files_only: bool = False,
        trust_remote_code: bool = False,
        task: Optional[str] = None,
        **kwargs,
    ) -> TemporaryDirectory:
        # Instantiate the transformers model checkpoint
        model = TasksManager.get_model_from_task(
            task=task,
            model_name_or_path=model_id,
            subfolder=subfolder,
            revision=revision,
            framework="pt",
            cache_dir=cache_dir,
            use_auth_token=use_auth_token,
            local_files_only=local_files_only,
            force_download=force_download,
            trust_remote_code=trust_remote_code,
            **kwargs,
        )

        # Save the model checkpoint in a temporary directory
        checkpoint_dir = TemporaryDirectory()
        model.save_pretrained(
            checkpoint_dir.name, save_function=save_split, safe_serialization=False, max_shard_size="10000GB"
        )
        return checkpoint_dir

    @classmethod
    @requires_transformers_neuronx
    def _from_transformers(
        cls,
        model_id: str,
        config: "PretrainedConfig",
        use_auth_token: Optional[str] = None,
        revision: Optional[str] = None,
        task: Optional[str] = None,
        batch_size: Optional[int] = 1,
        sequence_length: Optional[int] = None,
        num_cores: Optional[int] = 2,
        auto_cast_type: Optional[str] = "fp32",
        **kwargs,
    ) -> "NeuronDecoderModel":
        if task is None:
            task = TasksManager.infer_task_from_model(cls.auto_model_class)

        # Instantiate the transformers model checkpoint
        checkpoint_dir = cls._create_checkpoint(
            model_id,
            task=task,
            revision=revision,
            **kwargs,
        )

        if os.path.isdir(model_id):
            checkpoint_id = None
            checkpoint_revision = None
        else:
            checkpoint_id = model_id
            # Get the exact checkpoint revision (SHA1)
            api = HfApi(token=use_auth_token)
            model_info = api.repo_info(model_id, revision=revision)
            checkpoint_revision = model_info.sha

        # If the sequence_length was not specified, deduce it from the model configuration
        if sequence_length is None:
            # Note: for older models, max_position_embeddings is an alias for n_positions
            sequence_length = config.max_position_embeddings

        # Update the config
        config.neuron = {
            "task": task,
            "batch_size": batch_size,
            "num_cores": num_cores,
            "auto_cast_type": auto_cast_type,
            "sequence_length": sequence_length,
            "compiler_type": "neuronx-cc",
            "compiler_version": get_neuronxcc_version(),
            "checkpoint_id": checkpoint_id,
            "checkpoint_revision": checkpoint_revision,
        }

        # Try to reload the generation config (if any)
        generation_config = None
        try:
            generation_config = GenerationConfig.from_pretrained(model_id)
        except OSError:
            pass

        return cls(config, checkpoint_dir, generation_config=generation_config)

    @classmethod
    def _get_neuron_dirs(cls, model_path: Union[str, Path]) -> Tuple[str, str]:
        # The checkpoint is in a subdirectory
        checkpoint_dir = os.path.join(model_path, cls.CHECKPOINT_DIR)
        # So are the compiled artifacts
        compiled_dir = os.path.join(model_path, cls.COMPILED_DIR)
        return checkpoint_dir, compiled_dir

    @classmethod
    @requires_transformers_neuronx
    def _from_pretrained(
        cls,
        model_id: Union[str, Path],
        config: "PretrainedConfig",
        use_auth_token: Optional[str] = None,
        revision: Optional[str] = None,
        **kwargs,
    ) -> "NeuronDecoderModel":
        # Verify we are actually trying to load a neuron model
        neuron_config = getattr(config, "neuron", None)
        if neuron_config is None:
            raise ValueError(
                "The specified directory does not contain a neuron model."
                "Please convert your model to neuron format by passing export=True."
            )
        check_compiler_compatibility(neuron_config["compiler_type"], neuron_config["compiler_version"])

        model_path = model_id
        if not os.path.isdir(model_id):
            model_path = snapshot_download(model_id, token=use_auth_token, revision=revision)

        checkpoint_dir, compiled_dir = cls._get_neuron_dirs(model_path)
        if not os.path.isdir(checkpoint_dir):
            # Try to recreate checkpoint from neuron config
            task = neuron_config["task"]
            checkpoint_id = neuron_config.get("checkpoint_id", None)
            if checkpoint_id is None:
                raise ValueError("Unable to fetch the neuron model weights files.")
            checkpoint_revision = neuron_config["checkpoint_revision"]
            checkpoint_dir = cls._create_checkpoint(
                checkpoint_id,
                task=task,
                revision=checkpoint_revision,
                use_auth_token=use_auth_token,
                **kwargs,
            )
        assert os.path.isdir(compiled_dir)

        # Try to reload the generation config (if any)
        generation_config = None
        try:
            generation_config = GenerationConfig.from_pretrained(model_id)
        except OSError:
            pass

        return cls(config, checkpoint_dir, compiled_dir=compiled_dir, generation_config=generation_config)

    def forward(self, *args, **kwargs):
        raise NotImplementedError()

    def _save_pretrained(self, save_directory: Union[str, Path]):
        dst_checkpoint_path, dst_compiled_path = self._get_neuron_dirs(save_directory)

        def copy_dir_to_path(src_dir: Union[str, Path, TemporaryDirectory], dst_path: Union[str, Path]):
            if isinstance(src_dir, TemporaryDirectory):
                shutil.copytree(src_dir.name, dst_path)
            elif not os.path.samefile(src_dir, dst_path):
                os.symlink(dst_path, src_dir)

        # Copy checkpoint directory (it always exists)
        copy_dir_to_path(self.checkpoint_dir, dst_checkpoint_path)
        self.checkpoint_dir = dst_checkpoint_path
        # Save or create compiled directory
        if self.compiled_dir is None:
            # The compilation artifacts have never been saved, do it now
            self.model.save(dst_compiled_path)
        else:
            copy_dir_to_path(self.compiled_dir, dst_compiled_path)
        self.compiled_dir = dst_compiled_path
        self.generation_config.save_pretrained(save_directory)

    def push_to_hub(
        self,
        save_directory: str,
        repository_id: str,
        private: Optional[bool] = None,
        revision: Optional[str] = None,
        use_auth_token: Union[bool, str] = True,
        endpoint: Optional[str] = None,
    ) -> str:
        if isinstance(use_auth_token, str):
            huggingface_token = use_auth_token
        elif use_auth_token:
            huggingface_token = get_token()
        else:
            raise ValueError("You need to provide `use_auth_token` to be able to push to the hub")
        api = HfApi(endpoint=endpoint)

        user = api.whoami(huggingface_token)
        if is_google_colab():
            # Only in Google Colab to avoid the warning message
            self.git_config_username_and_email(git_email=user["email"], git_user=user["fullname"])

        api.create_repo(
            token=huggingface_token,
            repo_id=repository_id,
            exist_ok=True,
            private=private,
        )
        ignore_patterns = []
        neuron_config = getattr(self.config, "neuron")
        checkpoint_id = neuron_config.get("checkpoint_id", None)
        if checkpoint_id is not None:
            # Avoid uploading checkpoints when the original model is available on the hub
            ignore_patterns = [self.CHECKPOINT_DIR + "/*"]
        api.upload_folder(
            repo_id=repository_id,
            folder_path=save_directory,
            token=huggingface_token,
            revision=revision,
            ignore_patterns=ignore_patterns,
        )
